@@ -48,11 +48,13 @@ const char *const statusstrings[]= {
   N_("installed")
 };
 
-struct filenamenode *namenodetouse(struct filenamenode *namenode,
-				   struct pkginfo *pkg) {
+struct filenamenode *namenodetouse(struct filenamenode *namenode, struct pkginfo *pkg) {
   struct filenamenode *r;
   
-  if (!namenode->divert) { r= namenode; goto found; }
+  if (!namenode->divert) {
+    r = namenode;
+    goto found;
+  }
   
   debug(dbg_eachfile,"namenodetouse namenode=`%s' pkg=%s",
         namenode->name,pkg->name);
@@ -119,15 +121,6 @@ void checkpath(void) {
                 warned);
 }
 
-void ensure_package_clientdata(struct pkginfo *pkg) {
-  if (pkg->clientdata) return;
-  pkg->clientdata= nfmalloc(sizeof(struct perpackagestate));
-  pkg->clientdata->istobe= itb_normal;
-  pkg->clientdata->fileslistvalid= 0;
-  pkg->clientdata->files = NULL;
-  pkg->clientdata->trigprocdeferred= 0;
-}
-
 int ignore_depends(struct pkginfo *pkg) {
   struct packageinlist *id;
   for (id= ignoredependss; id; id= id->next)
@@ -160,6 +153,7 @@ static const char* preexecscript(const char *path, char *const *argv) {
 
   if (*instdir) {
     if (chroot(instdir)) ohshite(_("failed to chroot to `%.250s'"),instdir);
+    if (chdir("/")) ohshite(_("failed to chdir to `%.255s'"),"/");
   }
   if (f_debug & dbg_scripts) {
     fprintf(stderr,"D0%05o: fork/exec %s (",dbg_scripts,path);
@@ -179,7 +173,8 @@ static char *const *vbuildarglist(const char *scriptname, va_list ap) {
 
   i=0;
   if(bufs[0]) free(bufs[0]);
-  bufs[i++]= strdup(scriptname); /* yes, cast away const because exec wants it that way */
+  /* Yes, cast away const because exec wants it that way */
+  bufs[i++] = m_strdup(scriptname);
   for (;;) {
     assert(i < PKGSCRIPTMAXARGS);
     nextarg= va_arg(ap,char*);
@@ -231,18 +226,39 @@ static void script_catchsignals(void) {
   onerr_abort--;
 }
 
-void post_postinst_tasks(void) {
-  debug(dbg_triggersdetail,"post_postinst_tasks - trig_incorporate");
-  trig_incorporate(msdbrw_write,admindir);
-}
-static void post_script_tasks(void) {
-  debug(dbg_triggersdetail,"post_script_tasks -"
-	" ensure_diversions; trig_incorporate");
-  ensure_diversions();
-  trig_incorporate(msdbrw_write,admindir);
+void
+post_postinst_tasks(struct pkginfo *pkg, enum pkgstatus new_status)
+{
+  pkg->trigpend_head = NULL;
+  pkg->status = pkg->trigaw.head ? stat_triggersawaited : new_status;
+
+  post_postinst_tasks_core(pkg);
 }
 
-static void cu_post_script_tasks(int argc, void **argv){ post_script_tasks(); }
+void
+post_postinst_tasks_core(struct pkginfo *pkg)
+{
+  modstatdb_note(pkg);
+
+  debug(dbg_triggersdetail, "post_postinst_tasks_core - trig_incorporate");
+  trig_incorporate(msdbrw_write, admindir);
+}
+
+static void
+post_script_tasks(void)
+{
+  ensure_diversions();
+
+  debug(dbg_triggersdetail,
+        "post_script_tasks - ensure_diversions; trig_incorporate");
+  trig_incorporate(msdbrw_write, admindir);
+}
+
+static void
+cu_post_script_tasks(int argc, void **argv)
+{
+  post_script_tasks();
+}
 
 static void setexecute(const char *path, struct stat *stab) {
   if ((stab->st_mode & 0555) == 0555) return;
@@ -254,7 +270,7 @@ static int do_script(const char *pkg, const char *scriptname, const char *script
   int c1, r;
   setexecute(scriptpath,stab);
 
-  push_cleanup(cu_post_script_tasks,ehflag_bombout, 0,0, 0);
+  push_cleanup(cu_post_script_tasks, ehflag_bombout, NULL, 0, 0);
 
   c1= m_fork();
   if (!c1) {
@@ -265,8 +281,8 @@ static int do_script(const char *pkg, const char *scriptname, const char *script
       narglist[r]= arglist[r];
     scriptexec= preexecscript(scriptpath,(char * const *)narglist);
     narglist[0]= scriptexec;
-    if (setenv(MAINTSCRIPTPKGENVVAR,pkg,1) ||
-	setenv(MAINTSCRIPTDPKGENVVAR,PACKAGE_VERSION,1))
+    if (setenv(MAINTSCRIPTPKGENVVAR, pkg, 1) ||
+        setenv(MAINTSCRIPTDPKGENVVAR, PACKAGE_VERSION, 1))
       ohshite(_("unable to setenv for maint script"));
     execv(scriptexec,(char * const *)narglist);
     ohshite(desc,name);
@@ -276,13 +292,14 @@ static int do_script(const char *pkg, const char *scriptname, const char *script
   pop_cleanup(ehflag_normaltidy);
 
   pop_cleanup(ehflag_normaltidy);
+
   return r;
 }
 
-static int maintainer_script_installed_corev(struct pkginfo *pkg,
-					     const char *scriptname,
-					     const char *description,
-					     va_list ap) {
+static int
+vmaintainer_script_installed(struct pkginfo *pkg, const char *scriptname,
+                             const char *description, va_list ap)
+{
   const char *scriptpath;
   char *const *arglist;
   struct stat stab;
@@ -294,38 +311,49 @@ static int maintainer_script_installed_corev(struct pkginfo *pkg,
 
   if (stat(scriptpath,&stab)) {
     if (errno == ENOENT) {
-      debug(dbg_scripts,"maintainer_script_installed nonexistent %s",scriptname);
+      debug(dbg_scripts, "vmaintainer_script_installed nonexistent %s",
+            scriptname);
       return 0;
     }
     ohshite(_("unable to stat installed %s script `%.250s'"),description,scriptpath);
   }
   do_script(pkg->name, scriptname, scriptpath, &stab, arglist, _("unable to execute %s"), buf, 0);
+
   return 1;
 }
 
-int maintainer_script_installed(struct pkginfo *pkg, const char *scriptname,
-                                const char *description, ...) {
-  /* all ...'s are const char*'s */
+/* All ...'s are const char*'s. */
+int
+maintainer_script_installed(struct pkginfo *pkg, const char *scriptname,
+                            const char *description, ...)
+{
   int r;
   va_list ap;
-  va_start(ap,description);
-  r= maintainer_script_installed_corev(pkg, scriptname, description, ap);
+
+  va_start(ap, description);
+  r = vmaintainer_script_installed(pkg, scriptname, description, ap);
   va_end(ap);
-  if (r) post_script_tasks();
+  if (r)
+    post_script_tasks();
+
   return r;
 }
-  
-int maintainer_script_postinst(struct pkginfo *pkg, ...) {
+
+int
+maintainer_script_postinst(struct pkginfo *pkg, ...)
+{
   int r;
   va_list ap;
-  va_start(ap,pkg);
-  r= maintainer_script_installed_corev(pkg, POSTINSTFILE, "post-installation",
-				       ap);
+
+  va_start(ap, pkg);
+  r = vmaintainer_script_installed(pkg, POSTINSTFILE, "post-installation", ap);
   va_end(ap);
-  if (r) ensure_diversions();
+  if (r)
+    ensure_diversions();
+
   return r;
 }
-  
+
 int maintainer_script_new(const char *pkgname,
 			  const char *scriptname, const char *description,
                           const char *cidir, char *cidirrest, ...) {
@@ -349,6 +377,7 @@ int maintainer_script_new(const char *pkgname,
   }
   do_script(pkgname, scriptname, cidir, &stab, arglist, _("unable to execute new %s"), buf, 0);
   post_script_tasks();
+
   return 1;
 }
 
@@ -377,7 +406,8 @@ int maintainer_script_alternative(struct pkginfo *pkg,
             _("dpkg: warning - unable to stat %s `%.250s': %s\n"),
             buf,oldscriptpath,strerror(errno));
   } else {
-    if (!do_script(pkg->name, scriptname, oldscriptpath, &stab, arglist, _("unable to execute %s"), buf, PROCWARN)) {
+    if (!do_script(pkg->name, scriptname, oldscriptpath, &stab, arglist,
+                   _("unable to execute %s"), buf, PROCWARN)) {
       post_script_tasks();
       return 1;
     }
@@ -402,6 +432,7 @@ int maintainer_script_alternative(struct pkginfo *pkg,
   fprintf(stderr, _("dpkg: ... it looks like that went OK.\n"));
 
   post_script_tasks();
+
   return 1;
 }
 
@@ -415,6 +446,7 @@ void clear_istobes(void) {
     pkg->clientdata->istobe= itb_normal;
     pkg->clientdata->replacingfilesandsaid= 0;
   }
+  iterpkgend(it);
 }
 
 void debug(int which, const char *fmt, ...) {
@@ -540,21 +572,9 @@ void ensure_pathname_nonexisting(const char *pathname) {
   waitsubproc(c1,"rm cleanup",0);
 }
 
-int skip_due_to_hold(struct pkginfo *pkg) {
-  if (pkg->want != want_hold) return 0;
-  if (fc_hold) {
-    fprintf(stderr, _("Package %s was on hold, processing it anyway as you request\n"),
-            pkg->name);
-    return 0;
-  }
-  printf(_("Package %s is on hold, not touching it.  Use --force-hold to override.\n"),
-         pkg->name);
-  return 1;
-}
-
 void log_action(const char *action, struct pkginfo *pkg) {
   log_message("%s %s %s %s", action, pkg->name,
 	      versiondescribe(&pkg->installed.version, vdew_nonambig),
 	      versiondescribe(&pkg->available.version, vdew_nonambig));
-  statusfd_send("processing: %s: %s",action,pkg->name);
+  statusfd_send("processing: %s: %s", action, pkg->name);
 }
